extends "res://car/kinematic_vehicle.gd"

# majority of this is copy-pasted from the non-kinematic version
# I wish Godot could inherit from two scripts at once
var health = 100
var battery = 50

var World_node
var map

#hud
var hud
var speed_text
var minimap
var map_big
var panel
var game_over
var vjoy

var mouse_steer = false
var turn

var last_pos
var distance = 0
var distance_int = 0

# setup stuff
var elapsed_secs = 0
var start_secs = 2
var emitted = false

#signal load_ended

var chase_cam
var cockpit_cam = null
var debug_cam
# these are for cockpit cam
var cam_speed = 1
var cockpit_cam_target_angle = 0
var cockpit_cam_angle = 0
var cockpit_cam_max_angle = 5
var peek

# racing
var race
var prev = 0
var current = 0
var lap = 0
var dot = 0
var rel_loc = Vector3()
var race_path = PackedVector3Array()
var finished = false

var money = 0

# player navigation
var reached_inter
var reached_changed = false
var show_nav_tip = false

var hit = null
var count = 0
var was_tunnel = false # for particles
var was_dirt = false
var was_fast = false

var skidmark = null

# vehicle switching
var car_scene = null
var bike_scene = null

func _ready():
	# need to do it explicitly in Godot 4 for some reason
	super._ready()

	World_node = get_parent().get_parent().get_node("scene")
	cockpit_cam = $"cambase/CameraCockpit"
	debug_cam = $"cambase/CameraDebug"
	chase_cam = get_node(^"cambase/Camera3D")
	tail_mat = taillights.get_mesh().surface_get_material(0)

func on_load_ended():
	print("Loaded all pertinent stuff")
	# enable our cam
	chase_cam.make_current()
	# disable rear view mirror
	$"cambase/MirrorMesh".set_visible(false)
	$"cambase/SubViewport/CameraCockpitBack".clear_current()
	$"cambase/SubViewport".set_update_mode(SubViewport.UPDATE_DISABLED)

	# temporarily disable
	#get_node(^"driver_new").setup_ik()

	# optimize label/nameplate rendering
	get_node(^"..").freeze_viewports()

# ----------------------------------------------------------------
# kinematic driving
func get_input():
	turn = Input.get_action_strength("steer_left")
	turn -= Input.get_action_strength("steer_right")
	steer_target = turn * deg_to_rad(steering_limit)
	#var steer_angle = get_steering_angle(steer_target)
	#$tmpParent/Spatial_FL.rotation.y = steer_angle*2
	#$tmpParent/Spatial_FR.rotation.y = steer_angle*2
	acceleration = Vector3.ZERO
	if Input.is_action_pressed("accelerate"):
		acceleration = -transform.basis.z * engine_power
		#cancel braking visual
		if tail_mat != null:
			tail_mat.set_albedo(Color(0.62,0.62,0.62))
			tail_mat.set_feature(StandardMaterial3D.FEATURE_EMISSION, false)
	if Input.is_action_pressed("brake"):
		acceleration = -transform.basis.z * braking_power
		#visual effect
		if tail_mat != null:
			tail_mat.set_albedo(Color(1,1,1))
			tail_mat.set_feature(StandardMaterial3D.FEATURE_EMISSION, true)


# --------------------------------------------------
func damage_on_hit():
	if not get_parent().is_in_group("bike"):
		if speed > 5:
			print("Speed at collision: " + str(round(speed*3.6)) + "km/h, deducting: " + str(round(speed)))
			# deduct health
			health -= round(speed)

	if health <= 0:
		# game over!
		var over = game_over.instantiate()
		add_child(over)


func _physics_process(delta):
	# for some reason, needed in Godot 4
	super._physics_process(delta)


	# were we peeking last tick?
	var old_peek = peek

	# emit a signal when we're all set up
	elapsed_secs += delta
	if (elapsed_secs > start_secs and not emitted):
		#EventBus.emit_signal("load_ended")
		emitted = true

	# racing
	if race and race_path.size() > 0:
		var forward_global = get_global_transform() * (Vector3(0, 0, 2))
		var forward_vec = forward_global-get_global_transform().origin

		var pos = get_global_transform().origin
		#B-A = from A to B
		var target_vec = race_path[current] - pos
		# forward vec goes from origin to forward
		dot = forward_vec.dot(target_vec)

		rel_loc = race_path[current] * get_global_transform()

		#offset = offset_dist(race_path[prev], race_path[current], pos)

		position_on_line = position_line(prev, current, pos, race_path)

	# reset cam
	if turn == 0 and not peek:
		# if we were peeking last frame but are not now
		if old_peek and cockpit_cam.is_current():
			cockpit_cam_angle = 0
			#print("Old peek is: " + str(old_peek))

		# reset cam
		if cockpit_cam.is_current():
			cockpit_cam_target_angle = 0

	# vary cam speed

	cam_speed = 1
	if speed > 25:
		cam_speed = 4
	if speed > 15:
		cam_speed = 3
	else:
		cam_speed = 2

	if peek:
		cam_speed = 30

	#rotate cam
	# left
	if (cockpit_cam_target_angle < cockpit_cam_angle):
		cockpit_cam_angle -= cam_speed*delta
		if (cockpit_cam_target_angle > cockpit_cam_max_angle):
			#print("Setting to target angle: " + str(cockpit_cam_target_angle))
			cockpit_cam_angle = cockpit_cam_target_angle
	# right
	if (cockpit_cam_target_angle > cockpit_cam_angle):
		cockpit_cam_angle += cam_speed*delta
		#print("Cockpit cam angle: ", cockpit_cam_angle)
		# bugs
		#if (cockpit_cam.target_angle < cockpit_cam.max_angle):
		#	print("Setting to target angle: " + str(cockpit_cam.target_angle))
		#	cockpit_cam.angle = cockpit_cam.target_angle

	cockpit_cam.set_rotation(Vector3(deg_to_rad(180),deg_to_rad(cockpit_cam_angle), deg_to_rad(180)))

	# did we run into ...
	if get_slide_collision_count():
		var collision = get_slide_collision(0)
		# a cardboard box?
		if collision.get_collider() is RigidBody3D:
			var nam = collision.get_collider().get_name()
			#print(nam)
			# push on it
			var cr_imp = -get_global_transform().basis.z.normalized() * 4
			collision.get_collider().apply_impulse(Vector3(0,-2,0), Vector3(cr_imp.x, 0, cr_imp.z))
		# did we hit the train?
		if collision.get_collider() is AnimatableBody3D:
			# is it moving?
			if collision.get_collider().get_parent().get_node("AnimationPlayer").is_playing():
				# game over
				health = 0

		var nam = collision.get_collider().get_parent().get_name()
		#print(nam)
		# ignore ground or road "collisions"
		if "Ground" in nam or "Road" in nam:
			pass
		else:
			damage_on_hit()

func after_move():
	# racing ctd
	# track the path points
	if race and race_path.size() > 0:
		#if we passed the point, don't backtrack
		if (dot < 0 and rel_loc.distance_to(Vector3(0,0,0)) > 3 and rel_loc.distance_to(Vector3(0,0,0)) < 30):
			#print(get_parent().get_name() + " getting next point after " + str(current) + " because passed over " + str(dot))

			##do we have a next point?
			if (race_path.size() > current+1):
				prev = current
				current = current + 1
			else:
				if lap > 0:
					current = 0
					lap = lap+1
			#else:
				#print("We're at the end")
			#	stop = true

		if (rel_loc.distance_to(Vector3(0,0,0)) < 2):

			##do we have a next point?
			if (race_path.size() > current+1):
				#print("AI " + get_parent().get_name() + " gets a next point")
				prev = current
				current = current + 1
			else:
				if lap > 0:
					current = 0
					lap = lap+1

func reset_car():
	print("Reset!")
	#var tr = get_parent().get_translation()
	#get_parent().set_translation(Vector3(tr.x, 0.5, tr.z))
	#global_translate(Vector3(0, 0.5,0))
	translate_object_local(Vector3(0,0.1,0))
	# solution from https://godotengine.org/qa/56193/how-to-manually-set-the-position-of-a-kinematicbody2d
	set_velocity(Vector3(0,gravity/10,0))
	move_and_slide() #

	#gravity = 20

# UI stuff doesn't have to be in physics_process
func _process(delta):
	pass

#doesn't interact with physics
func _input(event):


	if (Input.is_action_pressed("camera_debug")):
		var chase_cam = get_node(^"cambase/Camera3D")
		if debug_cam.is_current():
			chase_cam.make_current()
#			# hud changes
#			hud.toggle_cam(true)
#			hud.speed_chase()
#			# show tunnels
#			var roads = get_tree().get_nodes_in_group("roads")
#			for r in roads:
#				if 'tunnel' in r and r.tunnel:
#					r.show_tunnel()
		else:
			debug_cam.make_current()
			# make tunnels transparent
			var roads = get_tree().get_nodes_in_group("roads")
			for r in roads:
				if 'tunnel' in r and r.tunnel:
					r.debug_tunnel()


	#reset
	if (Input.is_action_pressed("steer_reset")):
		reset_car()

# -----------------------------------------
# ---------------------------------

func get_compass_heading():
	# because E and W were easiest to identify (the sun)
	# this relied on Y rotation
	#var ang_to_dir = {180: "E", -180: "E", 0: "W", 90: "N", -90: "S"}
	# this relies on angle to marker
	var ang_to_dir = {180: "N", -180: "N", 0: "S", 90: "E", -90: "W"}

	# -180 -90 0 90 180 are the possible angles
	# this matches Y rot ang_to_dir above
	#var num_to_dir = {0: "E", 1:"S", 2:"W", 3:"N", 4:"E"}
	var num_to_dir = {0:"N", 1: "NW", 2:"W", 3: "SW", 4:"S", 5: "SE", 6:"E", 7: "NE", 8:"N"}
	# map from -180-180 to 0-4
	#var rot = get_rotation_degrees().y
	var rot = rad_to_deg(get_heading())
	var num_mapping = remap(rot, -180, 180, 0, 8)
	var disp = num_to_dir[int(round(num_mapping))]

	return disp

func get_heading():
	var forward_global = get_global_transform() * (Vector3(0, 0, -2))
	var forward_vec = forward_global-get_global_transform().origin
	#var basis_vec = player.get_global_transform().basis.z

	# looks like this is always positive?!
	#var player_rot = forward_vec.angle_to(Vector3(0,0,1))
	# returns radians
	#return player_rot
	if not has_node("/root/Node3D/marker_North"):
		return 0

	var North = get_node("/root/Node3D/marker_North")
	#if North == null:
	#	return 0
	var rel_loc = to_local(North.get_global_transform().origin)  #(North.get_global_transform() * get_global_transform().origin)
	#2D angle to target (local coords)
	var angle = atan2(rel_loc.x, rel_loc.z)
	#print("Heading: ", rad_to_deg(angle))
	return angle

# -----------------------------------------------
func create_race_path(path):
	print("Creating race path")
	if (path != null and path.size() > 0):
		#print("We have a race path to follow")
		for index in range(path.size()):
			race_path.push_back(path[index])

	# tell the race setup is done
	race.done = true
	print("Race set up is done")

func angle_to_intersection(id):
	# contains intersections global positions
	var intersections = get_node(^"Viewport_root/SubViewport/minimap").intersections
	var pos_gl = intersections[id]
	#print("Global position ", pos_gl)
	var rel_pos = pos_gl * get_global_transform()
	# dummy out the y value
	rel_pos = Vector3(rel_pos.x, 0, rel_pos.z)
	#print("Relative loc of intersection", id, " is ", rel_pos)
	# we don't care about z, only about x
	return rel_pos.x

# -----------------------------------
func swap_to_bike():
	# get positions
	var p_pos = get_parent().get_translation() #.get_global_transform().origin
	#print("Parent position before swap: " + str(p_pos))
	var pos = get_position()
	#print("Body pos before swap: " + str(pos))

	get_parent().queue_free()
	var bike = bike_scene.instantiate()
	bike.set_name("bike")
	# place the bike where the car was
	#bike.get_parent().global_transform.origin = p_pos
	bike.set_translation(p_pos)
	bike.get_node(^"BODY").set_translation(pos)

	get_parent().get_parent().add_child(bike)
	#print("Dummy")
	return bike.get_node(^"BODY")

func swap_to_car():
	# get positions
	var p_pos = get_parent().get_translation() #.get_global_transform().origin
	#print("Parent position before swap: " + str(p_pos))
	var pos = get_position()
	#print("Body pos before swap: " + str(pos))

	get_parent().queue_free()

	var car = car_scene.instantiate()
	car.set_name("car")
	# place the car where the bike was
	#car.get_parent().global_transform.origin = p_pos
	car.set_translation(p_pos)
	car.get_node(^"BODY").set_translation(pos)

	get_parent().get_parent().add_child(car)
	#print("Dummy")
	return car.get_node(^"BODY")


# -------------------------
func delay_new_events():
	var mmap = get_node(^"Viewport_root/SubViewport/minimap")
	mmap.add_event_markers()


func reset_events():
	var mmap = get_node(^"Viewport_root/SubViewport/minimap")
	var markers = get_tree().get_nodes_in_group("marker")
	# remove previous day's events
	for e in markers:
		e.queue_free()
		# remove minimap markers
		mmap.remove_marker(e.get_global_transform().origin)

	#TODO: add a message or a visual effect on the map!

	# new events
	var marker_data = map.get_node(^"nav").spawn_markers(map.samples, map.real_edges)
	#map.get_node(^"nav").setup_markers(marker_data)

	# we have to wait here because otherwise it shows markers for old events too
	# wait
	var t = Timer.new()
	t.set_wait_time(3)
	t.set_one_shot(true)
	self.add_child(t)
	t.start()
	await t.timeout
	# stuff after delay
	map.get_node(^"nav").setup_markers(marker_data)
	delay_new_events()
	t.queue_free()
